@@ -1,25 +1,13 @@
 /*  This file is part of the KDE project
 
-    Copyright(C) 2000 Alexander Neundorf <neundorf@kde.org>,
-        2014 Mathias Tillman <master.homer@gmail.com>
+    SPDX-FileCopyrightText: 2000 Alexander Neundorf <neundorf@kde.org>
+    SPDX-FileCopyrightText: 2014 Mathias Tillman <master.homer@gmail.com>
 
-    This library is free software; you can redistribute it and/or
-    modify it under the terms of the GNU Library General Public
-    License as published by the Free Software Foundation; either
-    version 2 of the License, or (at your option) any later version.
-
-    This library is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-    Library General Public License for more details.
-
-    You should have received a copy of the GNU Library General Public License
-    along with this library; see the file COPYING.LIB.  If not, write to
-    the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
-    Boston, MA 02110-1301, USA.
+    SPDX-License-Identifier: LGPL-2.0-or-later
 */
 
 #include "nfsv2.h"
+#include "kio_nfs_debug.h"
 
 #include <config-runtime.h>
 
@@ -46,7 +34,6 @@
 #include <QFile>
 #include <QDir>
 #include <QDebug>
-#include <QLoggingCategory>
 #include <QMimeDatabase>
 #include <QMimeType>
 
@@ -171,7 +158,7 @@ NFSFileHandle NFSProtocolV2::lookupFileHandle(const QString& path)
                                       clnt_timeout);
 
             if (rpcStatus == RPC_SUCCESS && readLinkRes.status == NFS_OK)
-            {						// get the absolute link target
+            {   // get the absolute link target
                 QString linkPath = QString::fromLocal8Bit(readLinkRes.readlinkres_u.data);
                 linkPath = QFileInfo(QFileInfo(path).path(), linkPath).absoluteFilePath();
 
@@ -258,7 +245,7 @@ void NFSProtocolV2::openConnection()
             // Save the exported directory and its NFS file handle.
             addFileHandle(fname, static_cast<NFSFileHandle>(fhStatus.fhstatus_u.fhs_fhandle));
             addExportedDir(fname);
-        } else {					// mount failed with error
+        } else {                    // mount failed with error
             qCDebug(LOG_KIO_NFS) << "Cannot mount" << fname << "- status" << fhStatus.fhs_status;
 
             // Even if the mount failed, record the directory path as exported
@@ -308,8 +295,8 @@ void NFSProtocolV2::listDir(const QUrl& url)
 {
     qCDebug(LOG_KIO_NFS) << url;
 
-    const QString path = listDirInternal(url);		// check path, list virtual dir
-    if (path.isEmpty()) return;				// no more to do
+    const QString path = listDirInternal(url);      // check path, list virtual dir
+    if (path.isEmpty()) return;             // no more to do
 
     const NFSFileHandle fh = getFileHandle(path);
     if (fh.isInvalid() || fh.isBadLink()) {
@@ -421,8 +408,8 @@ void NFSProtocolV2::stat(const QUrl &url)
 {
     qCDebug(LOG_KIO_NFS) << url;
 
-    const QString path = statInternal(url);		// check path, process virtual dir
-    if (path.isEmpty()) return;				// no more to do
+    const QString path = statInternal(url);     // check path, process virtual dir
+    if (path.isEmpty()) return;             // no more to do
 
     const NFSFileHandle fh = getFileHandle(path);
     if (fh.isInvalid())
