@@ -1,21 +1,8 @@
 /*  This file is part of the KDE libraries
-    Copyright (C) 2000 Malte Starostik <malte@kde.org>
-                  2000 Carsten Pfeiffer <pfeiffer@kde.org>
+    SPDX-FileCopyrightText: 2000 Malte Starostik <malte@kde.org>
+    SPDX-FileCopyrightText: 2000 Carsten Pfeiffer <pfeiffer@kde.org>
 
-    This library is free software; you can redistribute it and/or
-    modify it under the terms of the GNU Library General Public
-    License as published by the Free Software Foundation; either
-    version 2 of the License, or (at your option) any later version.
-
-    This library is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-    Library General Public License for more details.
-
-    You should have received a copy of the GNU Library General Public License
-    along with this library; see the file COPYING.LIB.  If not, write to
-    the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
-    Boston, MA 02110-1301, USA.
+    SPDX-License-Identifier: LGPL-2.0-or-later
 */
 
 #include "thumbnail.h"
@@ -23,7 +10,7 @@
 
 #include <stdlib.h>
 #ifdef __FreeBSD__
-    #include <machine/param.h>
+#include <machine/param.h>
 #endif
 #include <sys/types.h>
 #ifndef Q_OS_WIN
@@ -47,6 +34,7 @@
 #include <QMimeDatabase>
 #include <QLibrary>
 #include <QDebug>
+#include <QRandomGenerator>
 
 #include <KFileItem>
 #include <KLocalizedString>
@@ -56,9 +44,12 @@
 #include <KServiceTypeTrader>
 #include <KPluginLoader>
 
+#include <kio/previewjob.h>
 #include <kio/thumbcreator.h>
+#include <kio/thumbdevicepixelratiodependentcreator.h>
 #include <kio/thumbsequencecreator.h>
 #include <kio/previewjob.h>
+#include <kio_version.h>
 
 #include <limits>
 #include <QDirIterator>
@@ -81,6 +72,8 @@
 // plugin       - the name of the plugin library to be used for thumbnail creation.
 //                Provided by the application to save an addition KTrader
 //                query here.
+// devicePixelRatio - the devicePixelRatio to use for the output,
+//                     the dimensions of the output is multiplied by it and output pixmap will have devicePixelRatio
 // enabledPlugins - a list of enabled thumbnailer plugins. PreviewJob does not call
 //                  this thumbnail slave when a given plugin isn't enabled. However,
 //                  for directory thumbnails it doesn't know that the thumbnailer
@@ -96,6 +89,13 @@
 
 using namespace KIO;
 
+// Pseudo plugin class to embed meta data
+class KIOPluginForMetaData : public QObject
+{
+    Q_OBJECT
+    Q_PLUGIN_METADATA(IID "org.kde.kio.slave.thumbnail" FILE "thumbnail.json")
+};
+
 extern "C" Q_DECL_EXPORT int kdemain( int argc, char **argv )
 {
 #ifdef HAVE_NICE
@@ -129,21 +129,35 @@ extern "C" Q_DECL_EXPORT int kdemain( int argc, char **argv )
 
 ThumbnailProtocol::ThumbnailProtocol(const QByteArray &pool, const QByteArray &app)
     : SlaveBase("thumbnail", pool, app),
-      m_maxFileSize(0)
-{
-
-}
+      m_width(0),
+      m_height(0),
+      m_devicePixelRatio(1),
+      m_maxFileSize(0),
+      m_randomGenerator()
+{}
 
 ThumbnailProtocol::~ThumbnailProtocol()
 {
-    qDeleteAll( m_creators );
+    qDeleteAll(m_creators);
     m_creators.clear();
 }
 
+/**
+ * Scales down the image \p img in a way that it fits into the given maximum width and height
+ */
+void scaleDownImage(QImage& img, int maxWidth, int maxHeight)
+{
+    if (img.width() > maxWidth || img.height() > maxHeight) {
+        img = img.scaled(maxWidth,
+                         maxHeight,
+                         Qt::KeepAspectRatio, Qt::SmoothTransformation);
+    }
+}
+
 void ThumbnailProtocol::get(const QUrl &url)
 {
     m_mimeType = metaData("mimeType");
-    m_enabledPlugins = metaData("enabledPlugins").split(QLatin1Char(','), QString::SkipEmptyParts);
+    m_enabledPlugins = metaData("enabledPlugins").split(QLatin1Char(','), Qt::SkipEmptyParts);
     if (m_enabledPlugins.isEmpty()) {
         const KConfigGroup globalConfig(KSharedConfig::openConfig(), "PreviewSettings");
         m_enabledPlugins = globalConfig.readEntry("Plugins", KIO::PreviewJob::defaultPlugins());
@@ -154,13 +168,13 @@ void ThumbnailProtocol::get(const QUrl &url)
     Q_ASSERT(info.isAbsolute());
 
     if (!info.exists()) {
-	// The file does not exist
-	error(KIO::ERR_DOES_NOT_EXIST, url.path());
-	return;
+        // The file does not exist
+        error(KIO::ERR_DOES_NOT_EXIST, url.path());
+        return;
     } else if (!info.isReadable()) {
-	// The file is not readable!
-	error(KIO::ERR_CANNOT_READ, url.path());
-	return;
+        // The file is not readable!
+        error(KIO::ERR_CANNOT_READ, url.path());
+        return;
     }
 
     //qDebug() << "Wanting MIME Type:" << m_mimeType;
@@ -201,48 +215,61 @@ void ThumbnailProtocol::get(const QUrl &url)
         m_height = 128;
     }
 #endif
+    bool ok;
+    m_devicePixelRatio = metaData("devicePixelRatio").toInt(&ok);
+    if (!ok || m_devicePixelRatio == 0) {
+        m_devicePixelRatio = 1;
+    } else {
+        m_width *= m_devicePixelRatio;
+        m_height *= m_devicePixelRatio;
+    }
 
     QImage img;
-
-    KConfigGroup group( KSharedConfig::openConfig(), "PreviewSettings" );
-    bool kfmiThumb = false; // TODO Figure out if we can use KFileMetadata as a last resource
-
-    if (!kfmiThumb) {
-        QString plugin = metaData("plugin");
-        if ((plugin.isEmpty() || plugin == "directorythumbnail") && m_mimeType == "inode/directory") {
-            img = thumbForDirectory(info.canonicalFilePath());
-            if(img.isNull()) {
-              error(KIO::ERR_INTERNAL, i18n("Cannot create thumbnail for directory"));
-              return;
-            }
-        } else {
+    QString plugin = metaData("plugin");
+    if ((plugin.isEmpty() || plugin == "directorythumbnail") && m_mimeType == "inode/directory") {
+        img = thumbForDirectory(info.canonicalFilePath());
+        if (img.isNull()) {
+            error(KIO::ERR_INTERNAL, i18n("Cannot create thumbnail for directory"));
+            return;
+        }
+    } else {
 #ifdef THUMBNAIL_HACK
-            if (plugin.isEmpty()) {
-                plugin = pluginForMimeType(m_mimeType);
-            }
+        if (plugin.isEmpty()) {
+            plugin = pluginForMimeType(m_mimeType);
+        }
 
-            //qDebug() << "Guess plugin: " << plugin;
+        //qDebug() << "Guess plugin: " << plugin;
 #endif
-            if (plugin.isEmpty()) {
-                error(KIO::ERR_INTERNAL, i18n("No plugin specified."));
-                return;
-            }
+        if (plugin.isEmpty()) {
+            error(KIO::ERR_INTERNAL, i18n("No plugin specified."));
+            return;
+        }
 
-            ThumbCreator* creator = getThumbCreator(plugin);
-            if(!creator) {
-                error(KIO::ERR_INTERNAL, i18n("Cannot load ThumbCreator %1", plugin));
-                return;
-            }
+        ThumbCreatorWithMetadata* creator = getThumbCreator(plugin);
+        if (!creator) {
+            return;
+        }
 
-            ThumbSequenceCreator* sequenceCreator = dynamic_cast<ThumbSequenceCreator*>(creator);
-            if(sequenceCreator)
+        if (creator->handleSequences) {
+            ThumbSequenceCreator* sequenceCreator = dynamic_cast<ThumbSequenceCreator*>(creator->creator);
+            if (sequenceCreator) {
                 sequenceCreator->setSequenceIndex(sequenceIndex());
 
-            if (!creator->create(info.canonicalFilePath(), m_width, m_height, img)) {
-                error(KIO::ERR_INTERNAL, i18n("Cannot create thumbnail for %1", info.canonicalFilePath()));
-                return;
+                setMetaData("handlesSequences", QStringLiteral("1"));
             }
         }
+
+        if (!createThumbnail(creator, info.canonicalFilePath(), m_width, m_height, img)) {
+            error(KIO::ERR_INTERNAL, i18n("Cannot create thumbnail for %1", info.canonicalFilePath()));
+            return;
+        }
+
+        if (creator->handleSequences) {
+            ThumbSequenceCreator* sequenceCreator = dynamic_cast<ThumbSequenceCreator*>(creator->creator);
+            // We MUST do this after calling create(), because the create() call itself might change it.
+            const float wp = sequenceCreator->sequenceIndexWraparoundPoint();
+            setMetaData("sequenceIndexWraparoundPoint", QString().setNum(wp));
+        }
     }
 
     scaleDownImage(img, m_width, m_height);
@@ -292,13 +319,14 @@ void ThumbnailProtocol::get(const QUrl &url)
             img = img.convertToFormat(QImage::Format_ARGB32); //  so make sure there is none
         }
         struct shmid_ds shmStat;
-        if (shmctl(shmid.toInt(), IPC_STAT, &shmStat) == -1 || shmStat.shm_segsz < img.sizeInBytes()) {
+        if (shmctl(shmid.toInt(), IPC_STAT, &shmStat) == -1 || shmStat.shm_segsz < (uint)img.sizeInBytes()) {
             error(KIO::ERR_INTERNAL, i18n("Image is too big for the shared memory segment"));
             shmdt((char*)shmaddr);
             return;
         }
-        // Keep in sync with kdelibs/kio/kio/previewjob.cpp
-        stream << img.width() << img.height() << quint8(img.format());
+        // Keep in sync with kio/src/previewjob.cpp
+        const quint8 format = img.format() | 0x80;
+        stream << img.width() << img.height() << format << ((int)img.devicePixelRatio());
         memcpy(shmaddr, img.bits(), img.sizeInBytes());
         shmdt((char*)shmaddr);
         mimeType("application/octet-stream");
@@ -347,28 +375,30 @@ bool ThumbnailProtocol::isOpaque(const QImage &image) const
 }
 
 void ThumbnailProtocol::drawPictureFrame(QPainter *painter, const QPoint &centerPos,
-                                         const QImage &image, int frameWidth, QSize imageTargetSize) const
+        const QImage &image, int borderStrokeWidth, QSize imageTargetSize, int rotationAngle) const
 {
     // Scale the image down so it matches the aspect ratio
     float scaling = 1.0;
 
-    if ((image.size().width() > imageTargetSize.width()) && (imageTargetSize.width() != 0)) {
-        scaling = float(imageTargetSize.width()) / float(image.size().width());
-    } else if ((image.size().height() > imageTargetSize.height()) && (imageTargetSize.height() != 0)) {
-        scaling = float(imageTargetSize.height()) / float(image.size().height());
+    const bool landscapeDimension = image.width() > image.height();
+    const bool hasTargetSizeWidth = imageTargetSize.width() != 0;
+    const bool hasTargetSizeHeight = imageTargetSize.height() != 0;
+    const int widthWithFrames = image.width() + (2 * borderStrokeWidth);
+    const int heightWithFrames = image.height() + (2 * borderStrokeWidth);
+    if (landscapeDimension && (widthWithFrames > imageTargetSize.width()) && hasTargetSizeWidth) {
+        scaling = float(imageTargetSize.width()) / float(widthWithFrames);
+    } else if ((heightWithFrames > imageTargetSize.height()) && hasTargetSizeHeight) {
+        scaling = float(imageTargetSize.height()) / float(heightWithFrames);
     }
 
-    QImage frame(imageTargetSize + QSize(frameWidth * 2, frameWidth * 2),
-                 QImage::Format_ARGB32);
-    frame.fill(0);
-
-    float scaledFrameWidth = frameWidth / scaling;
+    const float scaledFrameWidth = borderStrokeWidth / scaling;
 
     QTransform m;
-    m.rotate(qrand() % 17 - 8); // Random rotation ±8°
+    m.rotate(rotationAngle);
     m.scale(scaling, scaling);
 
-    QRectF frameRect(QPointF(0, 0), QPointF(image.width() + scaledFrameWidth*2, image.height() + scaledFrameWidth*2));
+    const QRectF frameRect(QPointF(0, 0), QPointF(image.width() / image.devicePixelRatio() + scaledFrameWidth*2,
+                                                  image.height() / image.devicePixelRatio() + scaledFrameWidth*2));
 
     QRect r = m.mapRect(QRectF(frameRect)).toAlignedRect();
 
@@ -376,13 +406,13 @@ void ThumbnailProtocol::drawPictureFrame(QPainter *painter, const QPoint &center
     transformed.fill(0);
     QPainter p(&transformed);
     p.setRenderHint(QPainter::SmoothPixmapTransform);
+    p.setRenderHint(QPainter::Antialiasing);
     p.setCompositionMode(QPainter::CompositionMode_Source);
 
     p.translate(-r.topLeft());
     p.setWorldTransform(m, true);
 
     if (isOpaque(image)) {
-        p.setRenderHint(QPainter::Antialiasing);
         p.setPen(Qt::NoPen);
         p.setBrush(Qt::white);
         p.drawRoundedRect(frameRect, scaledFrameWidth / 2, scaledFrameWidth / 2);
@@ -390,7 +420,7 @@ void ThumbnailProtocol::drawPictureFrame(QPainter *painter, const QPoint &center
     p.drawImage(scaledFrameWidth, scaledFrameWidth, image);
     p.end();
 
-    int radius = qMax(frameWidth, 1);
+    int radius = qMax(borderStrokeWidth, 1);
 
     QImage shadow(r.size() + QSize(radius * 2, radius * 2), QImage::Format_ARGB32);
     shadow.fill(0);
@@ -414,12 +444,13 @@ QImage ThumbnailProtocol::thumbForDirectory(const QString& directory)
     if (m_propagationDirectories.isEmpty()) {
         // Directories that the directory preview will be propagated into if there is no direct sub-directories
         const KConfigGroup globalConfig(KSharedConfig::openConfig(), "PreviewSettings");
-        m_propagationDirectories = globalConfig.readEntry("PropagationDirectories", QStringList() << "VIDEO_TS").toSet();
+        const QStringList propagationDirectoriesList = globalConfig.readEntry("PropagationDirectories", QStringList() << "VIDEO_TS");
+        m_propagationDirectories = QSet<QString>(propagationDirectoriesList.begin(), propagationDirectoriesList.end());
         m_maxFileSize = globalConfig.readEntry("MaximumSize", std::numeric_limits<qint64>::max());
     }
 
     const int tiles = 2; //Count of items shown on each dimension
-    const int spacing = 1;
+    const int spacing = 1 * m_devicePixelRatio;
     const int visibleCount = tiles * tiles;
 
     // TODO: the margins are optimized for the Oxygen iconset
@@ -429,6 +460,7 @@ QImage ThumbnailProtocol::thumbForDirectory(const QString& directory)
     KFileItem item(QUrl::fromLocalFile(directory));
     const int extent = qMin(m_width, m_height);
     QPixmap folder = QIcon::fromTheme(item.iconName()).pixmap(extent);
+    folder.setDevicePixelRatio(m_devicePixelRatio);
 
     // Scale up base icon to ensure overlays are rendered with
     // the best quality possible even for low-res custom folder icons
@@ -443,10 +475,13 @@ QImage ThumbnailProtocol::thumbForDirectory(const QString& directory)
     const int bottomMargin = folderHeight / 6;
     const int leftMargin = folderWidth / 13;
     const int rightMargin = leftMargin;
+    // the picture border stroke width 1/170 rounded up
+    // (i.e for each 170px the folder width increases those border increase by 1 px)
+    const int borderStrokeWidth = qRound(folderWidth / 170.);
 
     const int segmentWidth  = (folderWidth  - leftMargin - rightMargin  + spacing) / tiles - spacing;
     const int segmentHeight = (folderHeight - topMargin  - bottomMargin + spacing) / tiles - spacing;
-    if ((segmentWidth < 5) || (segmentHeight <  5)) {
+    if ((segmentWidth < 5 * m_devicePixelRatio) || (segmentHeight < 5 * m_devicePixelRatio)) {
         // the segment size is too small for a useful preview
         return img;
     }
@@ -455,6 +490,7 @@ QImage ThumbnailProtocol::thumbForDirectory(const QString& directory)
     int skipValidItems = ((int)sequenceIndex()) * visibleCount;
 
     img = QImage(QSize(folderWidth, folderHeight), QImage::Format_ARGB32);
+    img.setDevicePixelRatio(m_devicePixelRatio);
     img.fill(0);
 
     QPainter p;
@@ -467,13 +503,12 @@ QImage ThumbnailProtocol::thumbForDirectory(const QString& directory)
     int xPos = leftMargin;
     int yPos = topMargin;
 
-    int frameWidth = qRound(folderWidth / 85.);
-
     int iterations = 0;
     QString hadFirstThumbnail;
     QImage firstThumbnail;
 
     int validThumbnails = 0;
+    int totalValidThumbs = -1;
 
     while (true) {
         QDirIterator dir(directory, QDir::Files | QDir::Readable);
@@ -481,8 +516,7 @@ QImage ThumbnailProtocol::thumbForDirectory(const QString& directory)
 
         // Seed the random number generator so that it always returns the same result
         // for the same directory and sequence-item
-        qsrand(qHash(directory) + skipValidItems);
-
+        m_randomGenerator.seed(qHash(directory) + skipValidItems);
         while (dir.hasNext()) {
             ++iterations;
             if (iterations > 500) {
@@ -517,9 +551,7 @@ QImage ThumbnailProtocol::thumbForDirectory(const QString& directory)
                 continue;
             }
 
-            if (!drawSubThumbnail(p, subThumbnail, segmentWidth, segmentHeight, xPos, yPos, frameWidth)) {
-                continue;
-            }
+            drawSubThumbnail(p, subThumbnail, segmentWidth, segmentHeight, xPos, yPos, borderStrokeWidth);
 
             if (hadFirstThumbnail.isEmpty()) {
                 hadFirstThumbnail = dir.filePath();
@@ -538,6 +570,12 @@ QImage ThumbnailProtocol::thumbForDirectory(const QString& directory)
             }
         }
 
+        if (!dir.hasNext() && totalValidThumbs < 0) {
+            // We iterated over the entire directory for the first time, so now we know how many thumbs
+            // were actually created.
+            totalValidThumbs = skipped+validThumbnails;
+        }
+
         if (validThumbnails > 0) {
             break;
         }
@@ -555,6 +593,14 @@ QImage ThumbnailProtocol::thumbForDirectory(const QString& directory)
 
     p.end();
 
+    if (totalValidThumbs >= 0) {
+        // We only know this once we've iterated over the entire directory, so this will only be
+        // set for large enough sequence indices.
+        const int wraparoundPoint = (totalValidThumbs-1)/visibleCount + 1;
+        setMetaData("sequenceIndexWraparoundPoint", QString().setNum(wraparoundPoint));
+    }
+    setMetaData("handlesSequences", QStringLiteral("1"));
+
     if (validThumbnails == 0) {
         // Eventually propagate the contained items from a sub-directory
         QDirIterator dir(directory, QDir::Dirs);
@@ -575,6 +621,7 @@ QImage ThumbnailProtocol::thumbForDirectory(const QString& directory)
     // If only for one file a thumbnail could be generated then paint an image with only one tile
     if (validThumbnails == 1) {
         QImage oneTileImg(folder.size(), QImage::Format_ARGB32);
+        oneTileImg.setDevicePixelRatio(m_devicePixelRatio);
         oneTileImg.fill(0);
 
         QPainter oneTilePainter(&oneTileImg);
@@ -588,43 +635,81 @@ QImage ThumbnailProtocol::thumbForDirectory(const QString& directory)
         if (firstThumbnail.width() < oneTileWidth && firstThumbnail.height() < oneTileHeight) {
             createSubThumbnail(firstThumbnail, hadFirstThumbnail, oneTileWidth, oneTileHeight);
         }
-        drawSubThumbnail(oneTilePainter, firstThumbnail, oneTileWidth, oneTileHeight, leftMargin, topMargin, frameWidth);
+        drawSubThumbnail(oneTilePainter, firstThumbnail, oneTileWidth, oneTileHeight, leftMargin, topMargin, borderStrokeWidth);
         return oneTileImg;
     }
 
     return img;
 }
 
-ThumbCreator* ThumbnailProtocol::getThumbCreator(const QString& plugin)
+ThumbCreatorWithMetadata* ThumbnailProtocol::getThumbCreator(const QString& plugin)
 {
     auto it = m_creators.constFind(plugin);
     if (it != m_creators.constEnd()) {
-	return *it;
+        return *it;
     }
 
-    ThumbCreator *creator = nullptr;
     // Don't use KPluginFactory here, this is not a QObject and
     // neither is ThumbCreator
+    ThumbCreator *creator = nullptr;
     QLibrary library(KPluginLoader::findPlugin((plugin)));
     if (library.load()) {
-	auto createFn = (newCreator)library.resolve("new_creator");
-	if (createFn) {
-	    creator = createFn();
-	}
-    }
-    if (!creator) {
-	qCWarning(KIO_THUMBNAIL_LOG) << "Failed to load" << plugin << library.errorString();
+        auto createFn = (newCreator)library.resolve("new_creator");
+        if (createFn) {
+            creator = createFn();
+        }
     }
 
-    m_creators.insert(plugin, creator);
+    ThumbCreatorWithMetadata *thumbCreator = nullptr;
+    if (creator) {
+        const KService::List plugins = KServiceTypeTrader::self()->query(QStringLiteral("ThumbCreator"), QStringLiteral("Library == '%1'").arg(plugin));
+        if (plugins.size() == 0) {
+            qCWarning(KIO_THUMBNAIL_LOG) << "Plugin not found:" << plugin;
+        } else {
+            auto service = plugins.first();
 
-    return creator;
+            QVariant cacheThumbnails = service->property("CacheThumbnail");
+            QVariant devicePixelRatioDependent = service->property("DevicePixelRatioDependent");
+            QVariant handleSequences = service->property("HandleSequences");
+
+            thumbCreator = new ThumbCreatorWithMetadata{
+                creator,
+                cacheThumbnails.isValid() ? cacheThumbnails.toBool() : true,
+                devicePixelRatioDependent.isValid() ? devicePixelRatioDependent.toBool() : false,
+                handleSequences.isValid() ? handleSequences.toBool() : false
+            };
+        }
+    } else {
+        qCWarning(KIO_THUMBNAIL_LOG) << "Failed to load" << plugin << library.errorString();
+    }
+
+    m_creators.insert(plugin, thumbCreator);
+
+    return thumbCreator;
 }
 
-bool ThumbnailProtocol::createSubThumbnail(QImage& thumbnail, const QString& filePath,
+void ThumbnailProtocol::ensureDirsCreated()
+{
+    if (m_thumbBasePath.isEmpty()) {
+        m_thumbBasePath = QStandardPaths::writableLocation(QStandardPaths::GenericCacheLocation) + QLatin1String("/thumbnails/");
+        QDir basePath(m_thumbBasePath);
+        basePath.mkpath("normal/");
+        QFile::setPermissions(basePath.absoluteFilePath("normal"), QFile::ReadOwner | QFile::WriteOwner | QFile::ExeOwner);
+        basePath.mkpath("large/");
+        QFile::setPermissions(basePath.absoluteFilePath("large"), QFile::ReadOwner | QFile::WriteOwner | QFile::ExeOwner);
+        if (m_devicePixelRatio > 1) {
+            basePath.mkpath("x-large/");
+            QFile::setPermissions(basePath.absoluteFilePath("x-large"), QFile::ReadOwner | QFile::WriteOwner | QFile::ExeOwner);
+            basePath.mkpath("xx-large/");
+            QFile::setPermissions(basePath.absoluteFilePath("xx-large"), QFile::ReadOwner | QFile::WriteOwner | QFile::ExeOwner);
+        }
+    }
+}
+
+bool ThumbnailProtocol::createSubThumbnail(QImage &thumbnail, const QString &filePath,
                                            int segmentWidth, int segmentHeight)
 {
-    auto getSubCreator = [&filePath, this]() -> ThumbCreator* {
+    auto getSubCreator = [&filePath, this]() -> ThumbCreatorWithMetadata* {
         const QMimeDatabase db;
         const QString subPlugin = pluginForMimeType(db.mimeTypeForFile(filePath).name());
         if (subPlugin.isEmpty() || !m_enabledPlugins.contains(subPlugin)) {
@@ -633,93 +718,142 @@ bool ThumbnailProtocol::createSubThumbnail(QImage& thumbnail, const QString& fil
         return getThumbCreator(subPlugin);
     };
 
-    if ((segmentWidth <= 256) && (segmentHeight <= 256)) {
+    const auto maxDimension = qMin(1024, 512 * m_devicePixelRatio);
+    if ((segmentWidth <= maxDimension) && (segmentHeight <= maxDimension)) {
         // check whether a cached version of the file is available for
-        // 128 x 128 or 256 x 256 pixels
+        // 128 x 128, 256 x 256 pixels or 512 x 512 pixels taking into account devicePixelRatio
         int cacheSize = 0;
         QCryptographicHash md5(QCryptographicHash::Md5);
         const QByteArray fileUrl = QUrl::fromLocalFile(filePath).toEncoded();
         md5.addData(fileUrl);
         const QString thumbName = QString::fromLatin1(md5.result().toHex()).append(".png");
 
-        if (m_thumbBasePath.isEmpty()) {
-            m_thumbBasePath = QStandardPaths::writableLocation(QStandardPaths::GenericCacheLocation) + QLatin1String("/thumbnails/");
-            QDir basePath(m_thumbBasePath);
-            basePath.mkpath("normal/");
-            QFile::setPermissions(basePath.absoluteFilePath("normal"), QFile::ReadOwner | QFile::WriteOwner | QFile::ExeOwner);
-            basePath.mkpath("large/");
-            QFile::setPermissions(basePath.absoluteFilePath("large"), QFile::ReadOwner | QFile::WriteOwner | QFile::ExeOwner);
-        }
+        ensureDirsCreated();
 
-        QDir thumbPath(m_thumbBasePath);
-        if ((segmentWidth <= 128) && (segmentHeight <= 128)) {
-            cacheSize = 128;
-            thumbPath.cd("normal");
-        } else {
-            cacheSize = 256;
-            thumbPath.cd("large");
-        }
+        struct CachePool {
+            QString path;
+            int minSize;
+        };
 
-        QFile thumbFile(thumbPath.absoluteFilePath(thumbName));
-        if (thumbFile.open(QIODevice::ReadOnly) && thumbnail.load(&thumbFile, "png")) {
-            return true;
-        } else if (cacheSize == 128) {
-            QDir fallbackPath(m_thumbBasePath);
-            fallbackPath.cd("large");
-            QFile fallbackThumbFile(fallbackPath.absoluteFilePath(thumbName));
-            if (fallbackThumbFile.open(QIODevice::ReadOnly) && thumbnail.load(&fallbackThumbFile, "png")) {
-                return true;
+        static const auto pools = {
+            CachePool{QStringLiteral("normal/"), 128},
+            CachePool{QStringLiteral("large/"), 256},
+            CachePool{QStringLiteral("x-large/"), 512},
+            CachePool{QStringLiteral("xx-large/"), 1024},
+        };
+
+        const int wants = std::max(segmentWidth, segmentHeight);
+        for (const auto &pool : pools) {
+            if (pool.minSize < wants) {
+                continue;
+            } else if (cacheSize == 0) {
+                // the lowest cache size the thumbnail could be at
+                cacheSize = pool.minSize;
+            }
+            // try in folders with higher image quality as well
+            if (thumbnail.load(m_thumbBasePath + pool.path + thumbName, "png")) {
+                thumbnail.setDevicePixelRatio(m_devicePixelRatio);
+                break;
             }
         }
 
         // no cached version is available, a new thumbnail must be created
-        ThumbCreator* subCreator = getSubCreator();
-        if (subCreator && subCreator->create(filePath, cacheSize, cacheSize, thumbnail)) {
-            scaleDownImage(thumbnail, cacheSize, cacheSize);
+        if (thumbnail.isNull()) {
+            ThumbCreatorWithMetadata* subCreator = getSubCreator();
+            if (subCreator && createThumbnail(subCreator, filePath, cacheSize, cacheSize, thumbnail)) {
+                scaleDownImage(thumbnail, cacheSize, cacheSize);
 
-            // The thumbnail has been created successfully. Store the thumbnail
-            // to the cache for future access.
-            QSaveFile thumbnailfile(thumbPath.absoluteFilePath(thumbName));
-            if (thumbnailfile.open(QIODevice::WriteOnly | QIODevice::Truncate)) {
-                QFileInfo fi(filePath);
-                thumbnail.setText(QStringLiteral("Thumb::URI"), QString::fromUtf8(fileUrl));
-                thumbnail.setText(QStringLiteral("Thumb::MTime"), QString::number(fi.lastModified().toSecsSinceEpoch()));
-                thumbnail.setText(QStringLiteral("Thumb::Size"), QString::number(fi.size()));
+                // The thumbnail has been created successfully. Check if we can store
+                // the thumbnail to the cache for future access.
+#if KIO_VERSION >= QT_VERSION_CHECK(5, 83, 0)
+                if (subCreator->cacheThumbnail && metaData("cache").toInt() && !thumbnail.isNull()) {
+#else
+                if (subCreator->cacheThumbnail && !thumbnail.isNull()) {
+#endif
+                    QString thumbPath;
+                    const int wants = std::max(thumbnail.width(), thumbnail.height());
+                    for (const auto &pool : pools) {
+                        if (pool.minSize < wants) {
+                            continue;
+                        } else if (thumbPath.isEmpty()) {
+                            // that's the appropriate path for this thumbnail
+                            thumbPath = m_thumbBasePath + pool.path;
+                        }
+                    }
 
-                if (thumbnail.save(&thumbnailfile, "png")) {
-                    thumbnailfile.commit();
+                    // The thumbnail has been created successfully. Store the thumbnail
+                    // to the cache for future access.
+                    QSaveFile thumbnailfile(QDir(thumbPath).absoluteFilePath(thumbName));
+                    if (thumbnailfile.open(QIODevice::WriteOnly | QIODevice::Truncate)) {
+                        QFileInfo fi(filePath);
+                        thumbnail.setText(QStringLiteral("Thumb::URI"), QString::fromUtf8(fileUrl));
+                        thumbnail.setText(QStringLiteral("Thumb::MTime"), QString::number(fi.lastModified().toSecsSinceEpoch()));
+                        thumbnail.setText(QStringLiteral("Thumb::Size"), QString::number(fi.size()));
+
+                        if (thumbnail.save(&thumbnailfile, "png")) {
+                            thumbnailfile.commit();
+                        }
+                    }
                 }
             }
-        } else {
+        }
+
+        if (thumbnail.isNull()) {
             return false;
         }
+
     } else {
-        ThumbCreator* subCreator = getSubCreator();
-        return subCreator && subCreator->create(filePath, segmentWidth, segmentHeight, thumbnail);
+        // image requested is too big to be stored in the cache
+        // create an image on demand
+        ThumbCreatorWithMetadata* subCreator = getSubCreator();
+        if (!subCreator || !createThumbnail(subCreator, filePath, segmentWidth, segmentHeight, thumbnail)) {
+            return false;
+        }
     }
+
+    // Make sure the image fits in the segments
+    // Some thumbnail creators do not respect the width / height parameters
+    scaleDownImage(thumbnail, segmentWidth, segmentHeight);
     return true;
 }
 
-void ThumbnailProtocol::scaleDownImage(QImage& img, int maxWidth, int maxHeight)
+bool ThumbnailProtocol::createThumbnail(ThumbCreatorWithMetadata* thumbCreator, const QString& filePath, int width, int height, QImage& thumbnail)
 {
-    if (img.width() > maxWidth || img.height() > maxHeight) {
-        img = img.scaled(maxWidth, maxHeight, Qt::KeepAspectRatio, Qt::SmoothTransformation);
+    int scaleWidth = width;
+    int scaledHeight = height;
+
+    if (thumbCreator->devicePixelRatioDependent) {
+        KIO::ThumbDevicePixelRatioDependentCreator *dprDependentCreator =
+                static_cast<KIO::ThumbDevicePixelRatioDependentCreator *>(
+                    thumbCreator->creator);
+
+        if (dprDependentCreator) {
+            dprDependentCreator->setDevicePixelRatio(m_devicePixelRatio);
+            scaleWidth /= m_devicePixelRatio;
+            scaledHeight /= m_devicePixelRatio;
+        }
     }
+
+    if (thumbCreator->creator->create(filePath, scaleWidth, scaledHeight, thumbnail)) {
+        // make sure the image is not bigger than the expected size
+        scaleDownImage(thumbnail, width, height);
+
+        thumbnail.setDevicePixelRatio(m_devicePixelRatio);
+
+        return true;
+    }
+
+    return false;
 }
 
-bool ThumbnailProtocol::drawSubThumbnail(QPainter& p, QImage subThumbnail, int width, int height, int xPos, int yPos, int frameWidth)
+void ThumbnailProtocol::drawSubThumbnail(QPainter& p, QImage subThumbnail, int width, int height, int xPos, int yPos, int borderStrokeWidth)
 {
-    // Apply fake smooth scaling, as seen on several blogs
-    if (subThumbnail.width() > width * 4 || subThumbnail.height() > height * 4) {
-        subThumbnail = subThumbnail.scaled(width*4, height*4, Qt::KeepAspectRatio, Qt::FastTransformation);
-    }
-
-    QSize targetSize(subThumbnail.size());
-    targetSize.scale(width, height, Qt::KeepAspectRatio);
+    scaleDownImage(subThumbnail, width, height);
 
     // center the image inside the segment boundaries
-    const QPoint centerPos(xPos + (width/ 2), yPos + (height / 2));
-    drawPictureFrame(&p, centerPos, subThumbnail, frameWidth, targetSize);
-
-    return true;
+    const QPoint centerPos((xPos + width/ 2) / m_devicePixelRatio, (yPos + height / 2) / m_devicePixelRatio);
+    const int rotationAngle = m_randomGenerator.bounded(-8, 9); // Random rotation ±8°
+    drawPictureFrame(&p, centerPos, subThumbnail, borderStrokeWidth, QSize(width, height), rotationAngle);
 }
+
+#include "thumbnail.moc"
